@@ -28,6 +28,89 @@ func _as_list($obj)
     return $obj;
 }
 
+# Check that all input names are in symbol's argument
+method _check_input_names(
+    AI::MXNet::Symbol $symbol,
+    ArrayRef[Str]     $names,
+    Str               $typename,
+    Bool              $throw
+)
+{
+    my @candidates;
+    my %args = map {
+        push @candidates, $_ if not /_(?:weight|bias|gamma|beta)$/;
+        $_ => 1
+    } @{ $symbol->list_arguments };
+    for my $name (@$names)
+    {
+        my $msg;
+        if(not exists $args{$name} and $name ne 'softmax_label')
+        {
+            $msg = sprintf("\033[91mYou created Module with Module(..., %s_names=%s) but "
+                ."input with name '%s' is not found in symbol.list_arguments(). "
+                ."Did you mean one of:\n\t%s\033[0m",
+                $typename, "@$names", $name, join("\n\t", @candidates)
+            );
+            if($throw)
+            {
+                confess($msg);
+            }
+            else
+            {
+                AI::MXNet::Logging->warning($msg);
+            }
+        }
+    }
+}
+
+# Check that input names matches input data descriptors
+method _check_names_match(
+    ArrayRef[Str]                  $data_names,
+    ArrayRef[NameShapeOrDataDesc]  $data_shapes,
+    Str                            $name,
+    Bool                           $throw
+)
+{
+    return if (not @$data_shapes and @$data_names == 1 and  $data_names->[0] eq 'softmax_label');
+    my @actual = map { @{$_}[0] } @{ $data_shapes };
+    if("@$data_names" ne "@actual")
+    {
+        my $msg = sprintf(
+            "Data provided by %s_shapes don't match names specified by %s_names (%s vs. %s)",
+            $name, $name, "@$data_shapes", "@$data_names"
+        );
+        if($throw)
+        {
+            confess($msg);
+        }
+        else
+        {
+            AI::MXNet::Logging->warning($msg);
+        }
+    }
+}
+
+method _parse_data_desc(
+    ArrayRef[Str]                                  $data_names,
+    Maybe[ArrayRef[Str]]                           $label_names,
+    ArrayRef[NameShapeOrDataDesc]                  $data_shapes,
+    Maybe[ArrayRef[NameShapeOrDataDesc]]           $label_shapes
+)
+{
+    $data_shapes = [map { blessed $_ ? $_ : AI::MXNet::DataDesc->new(@$_) } @$data_shapes];
+    $self->_check_names_match($data_names, $data_shapes, 'data', 1);
+    if($label_shapes)
+    {
+        $label_shapes = [map { blessed $_ ? $_ : AI::MXNet::DataDesc->new(@$_) } @$label_shapes];
+        $self->_check_names_match($label_names, $label_shapes, 'label', 0);
+    }
+    else
+    {
+        $self->_check_names_match($label_names, [], 'label', 0);
+    }
+    return ($data_shapes, $label_shapes);
+}
+
 =head1 DESCRIPTION
 
     The base class of a modules. A module represents a computation component. The design
@@ -112,12 +195,12 @@ func _as_list($obj)
     - `score`: run prediction on a data set and evaluate performance
 =cut
 
-has 'logger'  => (is => 'rw', default => sub { AI::MXNet::Logging->get_logger });
+has 'logger'            => (is => 'rw', default => sub { AI::MXNet::Logging->get_logger });
+has '_symbol'           => (is => 'rw', init_arg => 'symbol', isa => 'AI::MXNet::Symbol');
 has [
     qw/binded for_training inputs_need_grad
     params_initialized optimizer_initialized/
-]             => (is => 'rw', isa => 'Bool', init_arg => undef, default => 0);
-has '_symbol' => (is => 'rw', init_arg => 'symbol', isa => 'AI::MXNet::Symbol');
+]                       => (is => 'rw', isa => 'Bool', init_arg => undef, default => 0);
 
 ################################################################################
 # High Level API
@@ -246,7 +329,7 @@ method iter_predict(AI::MXNet::DataIter $eval_data, Maybe[Int] :$num_batch=, Boo
         $self->forward($eval_batch, is_train => 0);
         my $pad = $eval_batch->pad;
         my $outputs = [
-            map { $_->slice([0, $_->shape->[0] - $pad - 1]) } @{ $self->get_outputs() }
+            map { $_->slice([0, $_->shape->[0] - ($pad//0) - 1]) } @{ $self->get_outputs() }
         ];
         push @out, [$outputs, $nbatch, $eval_batch];
         $nbatch++;
@@ -302,7 +385,7 @@ method predict(
         last if defined $num_batch and $nbatch == $num_batch;
         $self->forward($eval_batch, is_train => 0);
         my $pad = $eval_batch->pad;
-        my $outputs = [map { $_->slice([0, $_->shape0->[0]-$pad-1])->copy } @{ $self->get_outputs }];
+        my $outputs = [map { $_->slice([0, $_->shape0->[0]-($pad//0)-1])->copy } @{ $self->get_outputs }];
         push @output_list, $outputs;
     }
     return () unless @output_list;
@@ -321,17 +404,17 @@ method predict(
         for my $i (0..$num_outputs-1)
         {
             push @output_list2,
-                 AI::MXNet::NDArray->concatenate(map { $_->[$i] } @output_list);
+                 AI::MXNet::NDArray->concatenate([map { $_->[$i] } @output_list]);
         }
         if($num_outputs == 1 and not $always_output_list)
         {
             return $output_list2[0];
         }
+        return @output_list2;
     }
     return @output_list;
 }
 
-
 =head2 fit
 
         Train the module parameters.
@@ -401,7 +484,7 @@ method fit(
     HashRef                            :$optimizer_params={ learning_rate => 0.01 },
     Maybe[Callback]                    :$eval_end_callback=,
     Maybe[Callback]                    :$eval_batch_end_callback=,
-    AI::MXNet::Initializer             :$initializer=AI::MXNet::Initializer->Uniform(learning_rate => 0.01),
+    AI::MXNet::Initializer             :$initializer=AI::MXNet::Initializer->Uniform(scale => 0.01),
     Maybe[HashRef[AI::MXNet::NDArray]] :$arg_params=,
     Maybe[HashRef[AI::MXNet::NDArray]] :$aux_params=,
     Bool                               :$allow_missing=0,
@@ -442,7 +525,6 @@ method fit(
     }
     $eval_metric = AI::MXNet::Metric->create($eval_metric)
         unless blessed $eval_metric;
-
     ################################################################################
     # training loop
     ################################################################################
@@ -458,7 +540,7 @@ method fit(
             $self->update;
             $self->update_metric($eval_metric, $data_batch->label);
             $monitor->toc_print if $monitor;
-            if($batch_end_callback)
+            if(defined $batch_end_callback)
             {
                 my $batch_end_params = AI::MXNet::BatchEndParam->new(
                     epoch       => $epoch,
@@ -470,6 +552,7 @@ method fit(
                     &{$callback}($batch_end_params);
                 }
             }
+            $nbatch++;
         }
         # one epoch of training is finished
         my $name_value = $eval_metric->get_name_value;
@@ -488,7 +571,7 @@ method fit(
         {
             for my $callback (@{ _as_list($epoch_end_callback) })
             {
-                &{$callback}($epoch, $self->symbol, $arg_params, $aux_params);
+                &{$callback}($epoch, $self->get_symbol, $arg_params, $aux_params);
             }
         }
         #----------------------------------------
@@ -517,6 +600,12 @@ method fit(
 # Symbol information
 ################################################################################
 
+=head2 get_symbol
+
+        Get the symbol used by this module.
+=cut
+method get_symbol() { $self->symbol }
+
 =head2 data_names
 
         A list of names for data required by this module.
@@ -644,9 +733,14 @@ method set_params(
             Path to output param file.
 =cut
 
-method save_params(Str $fname)
+method save_params(
+    Str $fname,
+    Maybe[HashRef[AI::MXNet::NDArray]] $arg_params=,
+    Maybe[HashRef[AI::MXNet::NDArray]] $aux_params=
+)
 {
-    my ($arg_params, $aux_params) = $self->get_params;
+    ($arg_params, $aux_params) = $self->get_params
+        unless (defined $arg_params and defined $aux_params);
     my %save_dict;
     while(my ($k, $v) = each %{ $arg_params })
     {
@@ -693,6 +787,52 @@ method load_params(Str $fname)
     $self->set_params(\%arg_params, \%aux_params);
 }
 
+=head2 get_states
+
+        Get states from all devices
+
+        Parameters
+        ----------
+        merge_multi_context : bool
+            Default is true (1). In the case when data-parallelism is used, the states
+            will be collected from multiple devices. A true value indicate that we
+            should merge the collected results so that they look like from a single
+            executor.
+
+        Returns
+        -------
+        If merge_multi_context is 1, it is like [out1, out2]. Otherwise, it
+        is like [[out1_dev1, out1_dev2], [out2_dev1, out2_dev2]]. All the output
+        elements are AI::MXNet::NDArray.
+=cut
+
+method get_states(Bool $merge_multi_context=1)
+{
+    assert($self->binded and $self->params_initialized);
+    assert(not $merge_multi_context);
+    return [];
+}
+
+=head2 set_states
+
+        Set value for states. Only one of states & value can be specified.
+
+        Parameters
+        ----------
+        states : list of list of NDArrays
+            source states arrays formatted like [[state1_dev1, state1_dev2],
+            [state2_dev1, state2_dev2]].
+        value : number
+            a single scalar value for all state arrays.
+=cut
+
+method set_states(Maybe[ArrayRef[ArrayRef[AI::MXNet::NDArray]]] $states=, Maybe[Num] $value=)
+{
+    assert($self->binded and $self->params_initialized);
+    assert(not $states and not $value);
+}
+
+
 =head2 install_monitor
 
         Install monitor on all executors
@@ -818,27 +958,27 @@ method update_metric(EvalMetric $eval_metric, ArrayRef[AI::MXNet::NDArray] $labe
 
         Parameters
         ----------
-        data_shapes : list of AI::MXNet::DataDesc
+        data_shapes : array ref of AI::MXNet::DataDesc
             Typically is `data_iter.provide_data`.
-        label_shapes : list of AI::MXNet::DataDesc
+        label_shapes : array ref of AI::MXNet::DataDesc
             Typically is `data_iter.provide_label`.
-        for_training : bool
-            Default is `True`. Whether the executors should be bind for training.
-        inputs_need_grad : bool
-            Default is `False`. Whether the gradients to the input data need to be computed.
+        for_training : Bool
+            Default is 1. Whether the executors should be bind for training.
+        inputs_need_grad : Bool
+            Default is 0. Whether the gradients to the input data need to be computed.
             Typically this is not needed. But this might be needed when implementing composition
             of modules.
-        force_rebind : bool
-            Default is `False`. This function does nothing if the executors are already
-            binded. But with this `True`, the executors will be forced to rebind.
+        force_rebind : Bool
+            Default is 0. This function does nothing if the executors are already
+            binded. But with this as 1, the executors will be forced to rebind.
         shared_module : Module
-            Default is `None`. This is used in bucketing. When not `None`, the shared module
+            Default is undef. This is used in bucketing. When not undef, the shared module
             essentially corresponds to a different bucket -- a module with different symbol
             but with the same sets of parameters (e.g. unrolled RNNs with different lengths).
-        grad_req : str, list of str, dict of str to str
+        grad_req : str, array ref of str, hashref of str to str
             Requirement for gradient accumulation. Can be 'write', 'add', or 'null'
             (default to 'write').
-            Can be specified globally (str) or for each argument (list, dict).
+            Can be specified globally (str) or for each argument (array ref, hash ref).
 =cut
 
 method bind(
